<?php

namespace Drupal\paragraphs_gantt\Plugin\Field\FieldFormatter;

use Drupal\Component\Serialization\Json;
use Drupal\Component\Utility\Html;
use Drupal\Core\Datetime\Entity\DateFormat;
use Drupal\Core\Entity\EntityDisplayRepositoryInterface;
use Drupal\Core\Field\FieldConfigInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\Plugin\Field\FieldFormatter\EntityReferenceFormatterBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Markup;
use Drupal\Core\Template\Attribute;
use Drupal\Core\Url;
use Drupal\field\Entity\FieldConfig;
use Drupal\paragraphs\Entity\Paragraph;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Plugin implementation of the paragraphs gantt formatter.
 *
 * @FieldFormatter(
 *   id = "paragraphs_gantt_formatter",
 *   module = "paragraphs_gantt",
 *   label = @Translation("Paragraphs gantt"),
 *   field_types = {
 *     "entity_reference_revisions"
 *   }
 * )
 */
class ParagraphsGanttFormatter extends EntityReferenceFormatterBase {

  /**
   * The entity display repository.
   *
   * @var \Drupal\Core\Entity\EntityDisplayRepositoryInterface
   */
  protected $entityDisplayRepository;

  /**
   * The fields of paragraphs.
   *
   * @var \Drupal\field\Entity\FieldConfig
   */
  protected $fieldsParagraphs;

  /**
   * Constructs a FormatterBase object.
   *
   * @param string $plugin_id
   *   The plugin_id for the formatter.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *   The definition of the field to which the formatter is associated.
   * @param array $settings
   *   The formatter settings.
   * @param string $label
   *   The formatter label display setting.
   * @param string $view_mode
   *   The view mode.
   * @param array $third_party_settings
   *   Any third party settings.
   * @param \Drupal\Core\Entity\EntityDisplayRepositoryInterface $entity_display_repository
   *   The entity display repository.
   */
  public function __construct($plugin_id, $plugin_definition, FieldDefinitionInterface $field_definition, array $settings, $label, $view_mode, array $third_party_settings, EntityDisplayRepositoryInterface $entity_display_repository) {
    parent::__construct($plugin_id, $plugin_definition, $field_definition, $settings, $label, $view_mode, $third_party_settings);
    $this->entityDisplayRepository = $entity_display_repository;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $plugin_id,
      $plugin_definition,
      $configuration['field_definition'],
      $configuration['settings'],
      $configuration['label'],
      $configuration['view_mode'],
      $configuration['third_party_settings'],
      $container->get('entity_display.repository')
    );
  }

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    // Implement default settings.
    return [
      'view_mode' => 'default',
      'text' => 'field_gantt_text',
      'type' => 'field_gantt_type',
      'start_date' => 'field_gantt_date',
      'duration' => 'field_gantt_duration',
      'progress' => 'field_gantt_progress',
      'open' => 'field_gantt_open',
      'parent' => 'field_gantt_parent',
      'links' => 'field_gantt_links',
      'owner' => 'field_gantt_owner',
      'cdn' => TRUE,
    ] + parent::defaultSettings();
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    $settingForm = [
      'view_mode' => [
        '#type' => 'select',
        '#options' => $this->entityDisplayRepository->getViewModeOptions($this->getFieldSetting('target_type')),
        '#title' => $this->t('View mode'),
        '#default_value' => $this->getSetting('view_mode'),
        '#required' => TRUE,
      ],
      'text' => [
        '#title' => $this->t('Title'),
        '#description' => $this->t('Field title'),
        '#type' => 'select',
        '#required' => TRUE,
        '#options' => $this->getConfigurableFields([
          'list_integer',
          'list_string',
          'text',
          'text_long',
          'text_with_summary',
          'string',
          'string_long',
        ]),
        '#empty_option' => t('- None -'),
        '#default_value' => $this->getSetting('text'),
      ],
      'start_date' => [
        '#title' => $this->t('Date'),
        '#description' => $this->t('Field date'),
        '#type' => 'select',
        '#required' => TRUE,
        '#options' => $this->getConfigurableFields([
          'datetime',
          'daterange',
        ]),
        '#empty_option' => t('- None -'),
        '#default_value' => $this->getSetting('start_date'),
      ],
      'type' => [
        '#title' => $this->t('Type'),
        '#description' => $this->t('Field type'),
        '#type' => 'select',
        '#required' => TRUE,
        '#options' => $this->getConfigurableFields([
          'list_string',
          'string',
        ]),
        '#empty_option' => t('- None -'),
        '#default_value' => $this->getSetting('type'),
      ],
      'duration' => [
        '#title' => $this->t('Duration'),
        '#description' => $this->t('Field duration'),
        '#type' => 'select',
        '#empty_option' => t('- None -'),
        '#options' => $this->getConfigurableFields([
          'integer',
        ]),
        '#default_value' => $this->getSetting('duration'),
      ],
      'progress' => [
        '#title' => $this->t('Progress'),
        '#description' => $this->t('Field progress'),
        '#type' => 'select',
        '#empty_option' => t('- None -'),
        '#options' => $this->getConfigurableFields([
          'float',
          'list_float',
          'decimal',
        ]),
        '#default_value' => $this->getSetting('progress'),
      ],
      'open' => [
        '#title' => $this->t('Open'),
        '#description' => $this->t('Field open'),
        '#type' => 'select',
        '#options' => $this->getConfigurableFields([
          'boolean',
        ]),
        '#default_value' => $this->getSetting('open'),
      ],
      'parent' => [
        '#title' => $this->t('Parent'),
        '#description' => $this->t('Field parent'),
        '#type' => 'select',
        '#empty_option' => t('- None -'),
        '#options' => $this->getConfigurableFields([
          'integer',
          'list_integer',
        ]),
        '#default_value' => $this->getSetting('parent'),
      ],
      'links' => [
        '#title' => $this->t('Links type'),
        '#description' => $this->t('Field link'),
        '#type' => 'select',
        '#empty_option' => t('- None -'),
        '#options' => $this->getConfigurableFields([
          'double_field',
        ]),
        '#default_value' => $this->getSetting('links'),
      ],
      'owner' => [
        '#title' => $this->t('Resource'),
        '#description' => $this->t('Resource Assignment'),
        '#type' => 'select',
        '#empty_option' => t('- None -'),
        '#options' => $this->getConfigurableFields([
          'integer',
          'entity_reference',
        ]),
        '#default_value' => $this->getSetting('owner'),
      ],
      'cdn' => [
        '#title' => $this->t('Use CDN'),
        '#description' => $this->t('If not you can use custom version in libraries'),
        '#type' => 'checkbox',
        '#default_value' => $this->getSetting('cdn'),
      ],
    ];
    return $settingForm + parent::settingsForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary() {
    if (empty($this->fieldsParagraphs)) {
      $this->setFieldsParagraphs();
    }
    $summary = parent::settingsSummary();
    if (!empty($this->getSetting('text'))) {
      $summary[] = $this->t('Field title: @field', ['@field' => $this->getLabelName($this->getSetting('text'))]);
    }
    if (!empty($this->getSetting('start_date'))) {
      $summary[] = $this->t('Field date: @field', ['@field' => $this->getLabelName($this->getSetting('start_date'))]);
    }
    if (!empty($this->getSetting('duration'))) {
      $summary[] = $this->t('Field duration: @field', ['@field' => $this->getLabelName($this->getSetting('duration'))]);
    }
    if (!empty($this->getSetting('progress'))) {
      $summary[] = $this->t('Field progress: @field', ['@field' => $this->getLabelName($this->getSetting('progress'))]);
    }
    if (!empty($this->getSetting('open'))) {
      $summary[] = $this->t('Field open: @field', ['@field' => $this->getLabelName($this->getSetting('open'))]);
    }
    if (!empty($this->getSetting('parent'))) {
      $summary[] = $this->t('Field parent: @field', ['@field' => $this->getLabelName($this->getSetting('parent'))]);
    }
    if (!empty($this->getSetting('owner'))) {
      $summary[] = $this->t('Field users: @field', ['@field' => $this->getLabelName($this->getSetting('owner'))]);
    }
    return $summary;
  }

  /**
   * Get list of fields.
   */
  protected function getConfigurableFields($type = FALSE) {
    if (empty($this->fieldsParagraphs)) {
      $this->setFieldsParagraphs();
    }
    $listField = [];
    foreach ($this->fieldsParagraphs as $field_name => $field) {
      $listField[$field_name] = $field->getLabel();
      if (!empty($type)) {
        if (!is_array($type) && $type != $field->getType()) {
          unset($listField[$field_name]);
        }
        if (is_array($type) && !in_array($field->getType(), $type)) {
          unset($listField[$field_name]);
        }
      }
    }
    return $listField;
  }

  /**
   * {@inheritdoc}
   */
  protected function setFieldsParagraphs() {
    $targetType = $this->getFieldSetting('target_type');
    $targetBundle = array_key_first($this->fieldDefinition->getSetting("handler_settings")["target_bundles"]);
    /** @var \Drupal\paragraphs\ParagraphInterface $paragraphs_entity */
    $paragraphs_entity = \Drupal::entityTypeManager()->getStorage($targetType)
      ->create(['type' => $targetBundle]);
    $field_definitions = $paragraphs_entity->getFieldDefinitions();
    foreach ($field_definitions as $field_name => $field) {
      if ($field instanceof FieldConfigInterface || $field instanceof FieldConfig) {
        $this->fieldsParagraphs[$field_name] = $field;
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  protected function getLabelName($fieldName) {
    if (!empty($this->fieldsParagraphs[$fieldName])) {
      return $this->fieldsParagraphs[$fieldName]->getLabel();
    }
    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function viewElements(FieldItemListInterface $items, $langcode) {

    $setting = $this->getSettings();
    $entity_type_id = $this->getFieldSetting('target_type');
    $field_definition = $items->getFieldDefinition();
    $selectionHandler = \Drupal::service('plugin.manager.entity_reference_selection')->getSelectionHandler($field_definition);
    $bundles = $selectionHandler->entityTypeBundleInfo->getBundleInfo($entity_type_id);
    $field_name = $field_definition->getName();
    $field_label = $field_definition->getLabel();
    $entity = $items->getEntity();
    $entityId = $entity->id();
    $currentURL = \Drupal::service('path.current')->getPath();
    $destination = ['destination' => $currentURL];
    $view_mode = $this->viewMode == 'full' ? 'default' : $this->viewMode;
    $dialog_width = '80%';
    $handlers = $field_definition->getSetting("handler_settings");
    $handler = array_key_first($handlers['target_bundles']);
    $storage = \Drupal::entityTypeManager()->getStorage('paragraphs_type');
    $paragraphs_type = $storage->load($handler);

    // Get list owner.
    $owner = [['key' => 0, 'label' => $this->t('- None -')]];
    if (!empty($setting['owner'])) {
      $paragraph = Paragraph::create(['type' => $handler]);
      $fieldDefinitions = $paragraph->getFieldDefinitions();
      $listOwners = \Drupal::service('plugin.manager.entity_reference_selection')
        ->getSelectionHandler($fieldDefinitions[$setting['owner']], $paragraph)
        ->getReferenceableEntities();
      if (!empty($listOwners['user'])) {
        $users = $listOwners['user'];
        if (!empty($users[0])) {
          unset($users[0]);
        }
        $colorIndex = 0;
        foreach ($users as $uid => $userName) {
          $owner[] = [
            'key' => $uid,
            'label' => $userName,
            'textColor' => "#FFF",
            'backgroundColor' => $this->generateColor($colorIndex++),
          ];
        }
      }
    }

    // Format date.
    $dateType = 'date';
    $setting['format_date'] = DateFormat::load('short')->getPattern();
    $paragraph = Paragraph::create(['type' => $paragraphs_type->id()]);
    $fieldDefinitions = $paragraph->getFieldDefinitions();
    if (!empty($fieldDefinitions[$setting['start_date']])) {
      $dateSettings = $fieldDefinitions[$setting['start_date']]->getSettings();
      $dateType = $dateSettings["datetime_type"];
    }
    if ($dateType == 'date') {
      $setting['format_date'] = current(explode(' ', $setting['format_date']));
    }

    $path = '';
    if (method_exists($paragraphs_type, 'getIconUrl')) {
      $path = $paragraphs_type->getIconUrl();
    }
    $title = implode(' ', [
      '<i class="bi bi-calendar-plus"></i>',
      $this->t('Add'),
      $bundles[$handler]['label'],
    ]);
    if (!empty($path)) {
      $title = '<img src="' . $path . '" height="25"/> ' . $title;
      $field_label = '<img src="' . $path . '" height="25"/> ' . $field_label;
    }
    $ganttId = $handler . '_' . $entityId;
    $attr = [
      'class' => ['gantt', $handler],
      'id' => Html::getId($ganttId),
    ];
    $lib = 'paragraphs_gantt/paragraphs-gantt';
    if ($setting['cdn']) {
      $lib .= '.cdn';
    }
    $elements = [
      '#type' => 'container',
      '#theme' => 'paragraphs_gantt_wrapper',
      '#field_name' => $field_name,
      '#title' => $field_label,
      '#id' => $ganttId,
      '#owner' => $owner,
      '#data' => $this->convertGanttData($items, $langcode, $setting),
      '#attributes' => new Attribute($attr),
      '#ajax' => Url::fromRoute('paragraphs.gantt', [
        'paragraph_type' => $handler,
        'entity_type' => $entity_type_id,
        'entity_field' => $field_name,
        'entity_id' => $entityId,
      ]),
      '#add' => [
        '#type' => 'link',
        '#url' => Url::fromRoute('paragraphs.add', [
          'paragraph_type' => $handler,
          'entity_type' => $entity_type_id,
          'entity_field' => $field_name,
          'entity_id' => $entityId,
        ])->setOptions(['query' => $destination]),
        '#title' => Markup::create($title),
        '#attributes' => [
          'class' => ['use-ajax' , 'btn' , 'btn-default'],
          'data-dialog-type' => 'modal',
          'data-dialog-options' => Json::encode(['width' => $dialog_width]),
        ],
      ],
      '#settings' => $setting,
      '#attached' => [
        'drupalSettings' => [
          'date_format' => str_replace(
            ['Y', 'm', 'd', 'H', 'i', 's'],
            ['%Y', '%m', '%d', '%H', '%i', '%s'],
            $setting['format_date']
          ),
          'gantt' => [
            $handler => [
              'id' => 'gantt_' . $ganttId,
              'owner' => $setting['owner'],
              'ajax' => Url::fromRoute('paragraphs.gantt', [
                'paragraph_type' => $handler,
                'view_mode' => $view_mode,
                'entity_type' => $entity_type_id,
                'entity_field' => $field_name,
                'entity_id' => $entityId,
              ])->toString(),
              'import' => Url::fromRoute('paragraphs.import', [
                'paragraph_type' => $handler,
                'view_mode' => $view_mode,
                'entity_type' => $entity_type_id,
                'entity_field' => $field_name,
                'entity_id' => $entityId,
              ])->toString(),
            ],
          ],
        ],
        'library' => [$lib],
      ],
    ];
    return $elements;
  }

  /**
   * {@inheritdoc}
   */
  protected function convertGanttData(FieldItemListInterface $items, string $langcode, array $setting) {

    $entities = $this->getEntitiesToView($items, $langcode);
    // When a display renderer doesn't exist, fall back to the default.
    $field_definition = $items->getFieldDefinition();
    $handlers = $field_definition->getSetting("handler_settings");
    $targetBundle = array_key_first($handlers['target_bundles']);
    $links = [];
    $outData = [];
    $mappingGantt = [
      'text',
      'type',
      'start_date',
      'duration',
      'progress',
      'open',
      'parent',
      'owner',
    ];
    $fieldDateType = 'date';
    $fieldDefinitions = \Drupal::service('entity_field.manager')
      ->getFieldDefinitions($field_definition->getTargetEntityTypeId(), $targetBundle);
    if (!empty($fieldDefinitions[$setting['start_date']])) {
      $dateSettings = $fieldDefinitions[$setting['start_date']]->getSettings();
      $fieldDateType = $fieldDefinitions[$setting['start_date']]->getType();
    }

    foreach ($entities as $delta => $entity) {
      if ($targetBundle != $entity->bundle()) {
        continue;
      }

      $data[$delta] = ['id' => $entity->id()];
      foreach ($mappingGantt as $ganttField) {
        if (!empty($field_name = $setting[$ganttField])) {
          $value = $entity->get($field_name)->getString();
          if (!empty($value)) {
            $data[$delta][$ganttField] = $value;
          }
          if ($ganttField == 'start_date') {
            if (!empty($entity->get($field_name)->date)) {
              $data[$delta][$ganttField] = $entity->get($field_name)->date->format($setting['format_date']);
            }
            if ($fieldDateType == 'daterange') {
              $dateValue = $entity->get($field_name)->value;
              // $endValue = $entity->get($field_name)->end_value;
              $data[$delta][$ganttField] = \Drupal::service('date.formatter')
                ->format(strtotime($dateValue), 'short');
            }
          }

          // Process owner.
          if ($ganttField == 'owner' && !$entity->get($field_name)->isEmpty()) {
            $values = $entity->get($field_name)->getValue();
            $data[$delta][$ganttField] = [];
            foreach ($values as $value) {
              if (!empty($value->entity)) {
                $data[$delta][$ganttField][] = $value->entity->id();
              }
            }
            // @todo remove later for multiuser.
            $data[$delta][$ganttField] = $value->entity->id();
          }
        }
      }
      // Process Links.
      if (!empty($setting['links'])) {
        $valueLinks = $entity->get($setting['links'])->getValue();
        foreach ($valueLinks as $valueLink) {
          $links[] = [
            'id' => implode('-', [$entity->id(), $valueLink['first']]),
            'source' => $entity->id(),
            'target' => (int) $valueLink['first'],
            'type' => (int) $valueLink['second'],
          ];
        }
      }

    }
    if (!empty($data)) {
      $outData['data'] = $data;
      if (!empty($links)) {
        $outData['links'] = $links;
      }
    }
    return $outData;
  }

  /**
   * Change color backgroup foreach user.
   *
   * @param int $index
   *   Index user in array.
   *
   * @return string
   *   Code color.
   */
  private function generateColor(int $index) {
    $listColor = ["#03A9F4", "#f57730", "#e157de", "#78909C", "#8D6E63",
      "#800000", "#008080", "#d2691e", "#ff8c00", "#daa520", "#4b0082",
      "#FF0000", "#800080", "#008000", "#808000", "#000080", "#0000FF",
    ];
    if (empty($listColor[$index])) {
      return $listColor[$index % count($listColor)];
    }
    return $listColor[$index];
  }

}
